/* arch/arm/mach-s3c2410/tomtomgo-wake.c
 *
 * Definitions of TomTom GO wakeup check functions.
 *
 * Copyright (C) 2005 TomTom BV <http://www.tomtom.com/>
 * Authors: Dimitry Andric <dimitry.andric@tomtom.com>
 *          Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-adc.h>
#include <asm/arch/system.h>
#include <asm/arch/tomtomgo-wake.h>
#include <barcelona/gopins.h>
#include "tomtomgo-iopins.h"

#define BOOTLOADER_START_ADDRESS 0x8024
#define SIMULATED_WAKE 0xa5a5a5a5

#ifdef CONFIG_S3C2410_PM_DEBUG
extern void printascii(const char *);
#define DBG(msg) printascii("tomtomgo-wake: " msg "\n")
#else /* CONFIG_S3C2410_PM_DEBUG */
#define DBG(msg) do {} while (0)
#endif /* CONFIG_S3C2410_PM_DEBUG */

static int tomtomgo_decide_battery(void);

void tomtomgo_before_sleep(int simulateWake, int linuxSuspend)
{
#ifdef CONFIG_S3C2410_PM_DEBUG
	/* Wait a little while to flush debug messages */
	mdelay(10);
#endif /* CONFIG_S3C2410_PM_DEBUG */
	// Disable dock power
	IO_Deactivate(DOCK_PWREN);
	// Wait a while so dockint gets stable
	mdelay(100);
	// Disable all IO to be sure to enter low power mode
	IO_PowerOff();
	// Set wakeup sources
	if (simulateWake)
	{
		// Simulate direct wakeup by using non pressed power button
		IO_SetInterruptOnDeactivated(ON_OFF);
		// Set resume address to bootloader
		__raw_writel(BOOTLOADER_START_ADDRESS, S3C2410_GSTATUS3);
		// Bootloader checks GSTATUS4 for simulated wake
		__raw_writel(SIMULATED_WAKE, S3C2410_GSTATUS4);
	}
	else
	{
		// Set power button as wakeup source
		IO_SetInterruptOnActivation(ON_OFF);
		// Only set other wakeupsources when staying in Linux
		if (linuxSuspend)
		{
			// Set ignition as wakeup source
			IO_SetInterruptOnActivation(IGNITION);
			// Set SD Card removal as wakeup source
			IO_SetInterruptOnDeactivation(CD_SD);
			// Set USB cable insertion as wakeup source
			IO_SetInterruptOnActivation(USB_HOST_DETECT);
			// Set docking interrupt as wakeup source
			IO_SetInterruptOnActivation(DOCK_INT);
			// Resume address of Linux is/was set by pm.c
		}
		else
		{
			// Suspend to bootloader, so set resume address to bootloader
			__raw_writel(BOOTLOADER_START_ADDRESS, S3C2410_GSTATUS3);
			// Bootloader checks GSTATUS4 for simulated wake
			__raw_writel(0, S3C2410_GSTATUS4);
		}
	}
}

int tomtomgo_decide_wakeup(void)
{
	// Enable dock power so IO status can be read
	IO_Activate(DOCK_PWREN);
	// Wait a while for stable IO
	mdelay(50);
	// Update IO state
	IO_Update(1);

#if 0 // Do not reboot if USB is inserted, because the application will now ask for it  
	if (IO_GetInput(USB_HOST_DETECT)) 
	{
		DBG("USB inserted, suspending to bootloader");
		goto bootloader_suspend;
	}
#endif // 0

	// Goto sleep again with resume to bootloader when SD card removed
	DBG("checking SD card removal");
	if (IO_HaveSdCard() && !IO_GetInput(CD_SD)) 
	{
		DBG("SD card removed, suspending to bootloader");
		goto bootloader_suspend;
	}
#if 0
	// Reboot to bootloader when HDD temperature is out of range
	DBG("checking HDD temp");
	if (IO_HaveHarddisk() && tomtom_hddtemp_check() < 0) 
	{
		DBG("HDD temp bad, suspending to bootloader");
		goto bootloader_suspend;
	}
#endif

	// Goto sleep again if battery is too low
	DBG("checking battery");
	if (!tomtomgo_decide_battery()) 
	{
		DBG("battery too low, suspending to linux");
		goto linux_suspend;
	}

	// Check docking state
	DBG("checking dock sense");
	if ((IO_GetDockState() == GODOCK_NONE) || (IO_GetDockState() == GODOCK_DESK))
	{
		// Not docked so check power button
		int i = IO_HaveLoweredButton() ? 100 : 1500;
		// Power button should be pressed longer for classic GO
		DBG("checking power button");
		while (i--) 
		{
			mdelay(1);
			if (!IO_GetInput(ON_OFF)) 
			{
				DBG("power button bounce, suspending to linux");
				goto linux_suspend;
			}
		}
	}
	else
	{
		// Docked so check power button and ignition
		if ((!IO_GetInput(IGNITION)) && (!IO_GetInput(ON_OFF))) goto linux_suspend;
	}
	DBG("going to wake up");
	return 1;

bootloader_suspend:
	tomtomgo_before_sleep(0, 0);
	return 0;

linux_suspend:
	tomtomgo_before_sleep(0, 1);
	return 0;
}


/********************************************************************
 * Battery check
 ********************************************************************/

#define BATTERY_MIN_VALUE 3650

static int tomtomgo_decide_battery(void)
{
	struct clk *clk;
	int i;
	unsigned long adctsc, adccon;
	unsigned long my_adctsc, my_adccon;
	unsigned long battery, reference, voltage;

	/* cannot check battery when charging... */
	if (IO_GetInput(ACPWR)) return 1;

	clk = clk_get(NULL, "adc");
	if (!clk)
		return 1;

	/* Save touch screen & ADC registers */
	adctsc = __raw_readl(S3C2410_ADCTSC);
	adccon = __raw_readl(S3C2410_ADCCON);

	/* Activate reference voltage source */
	IO_Activate(AIN4_PWR);

	/* Setup adc */
	my_adccon = (0xff << 6) | (1 << 14);
	my_adctsc = (1 << 3) | (1 << 4) | (1 << 6);

	/* Enable clock and let it settle a bit */
	clk_enable(clk);
	mdelay(10);

	/* Write my values */
	__raw_writel(my_adctsc, S3C2410_ADCTSC);
	__raw_writel(my_adccon, S3C2410_ADCCON);

	/* Battery and reference voltages are averaged over 10 samples. */
	battery = reference = 0;

	for(i = 0; i < 10; i++) 
	{
		/* Select battery voltage channel and wait a while */
		__raw_writel(my_adccon | 0 | (0 << 3), S3C2410_ADCCON);
		mdelay(5);
		/* Start conversion of battery voltage */
		__raw_writel(my_adccon | 1 | (0 << 3), S3C2410_ADCCON);

		/* Wait for start of and end of conversion */
		/* TODO: Timeout? */
		while (__raw_readl(S3C2410_ADCCON) & 1)
			;
		while (!(__raw_readl(S3C2410_ADCCON) & (1<<15)))
			;
		/* Read result; */
		battery+= __raw_readl(S3C2410_ADCDAT0) & 0x3FF;

		/* Select reference voltage and wait a while */
		__raw_writel(my_adccon | 0 | (4 << 3), S3C2410_ADCCON);
		mdelay(5);
		/* Start conversion of reference voltage */
		__raw_writel(my_adccon | 1 | (4 << 3), S3C2410_ADCCON);

		/* Wait for start of and end of conversion */
		/* TODO: Timeout? */
		while (__raw_readl(S3C2410_ADCCON) & 1)
			;
		while (!(__raw_readl(S3C2410_ADCCON) & (1<<15)))
			;
		/* Read result; */
		reference+= __raw_readl(S3C2410_ADCDAT0) & 0x3FF;
	}

	battery /= 10;
	reference /= 10;

	/* Calculate real battery voltage */
	voltage = 2450 * battery;
	voltage /= reference;

	/* disable clock and restore original register contents */
	__raw_writel(adccon, S3C2410_ADCCON);
	__raw_writel(adctsc, S3C2410_ADCTSC);

	clk_disable(clk);
	clk_put(clk);

	IO_Deactivate(AIN4_PWR);

	return (voltage >= BATTERY_MIN_VALUE) ? 1 : 0;
}

/********************************************************************
 * IDE check
 ********************************************************************/

static unsigned long ideBaseAddress;

// IDE interface
#define rIDEDATAREAD     (*(volatile unsigned short*)(ideBaseAddress+(0*4)))
#define rIDEDATAWRITE    (*(volatile unsigned short*)(ideBaseAddress+(0*4)))
#define rIDEERROR        (*(volatile unsigned char*) (ideBaseAddress+(1*4)))
#define rIDEFEATURES     (*(volatile unsigned char*) (ideBaseAddress+(1*4)))
#define rIDESECTORCOUNT  (*(volatile unsigned char*) (ideBaseAddress+(2*4)))
#define rIDESECTORNUMBER (*(volatile unsigned char*) (ideBaseAddress+(3*4)))
#define rIDECYLINDERLOW  (*(volatile unsigned char*) (ideBaseAddress+(4*4)))
#define rIDECYLINDERHIGH (*(volatile unsigned char*) (ideBaseAddress+(5*4)))
#define rIDEDEVICEHEAD   (*(volatile unsigned char*) (ideBaseAddress+(6*4)))
#define rIDESTATUS       (*(volatile unsigned char*) (ideBaseAddress+(7*4)))
#define rIDECOMMAND      (*(volatile unsigned char*) (ideBaseAddress+(7*4)))

/* bits in Status register: */
#define BSY 0x0080

static void IDE_WaitForBusyCleared(void)
{
	while ((rIDESTATUS & BSY) > 0);
}

static void IDE_DoCommand(unsigned short cCommand)
{
	IDE_WaitForBusyCleared();
	rIDECOMMAND = cCommand;
	IDE_WaitForBusyCleared();
}

static void IDE_InitDisk(void)
{
	IO_SetInput(HDD_DRQ);
	IO_SetInput(HDD_DACK);
	IO_Activate(HDD_PWR_ON);
	IO_SetInput(HDD_IRQ);
	IO_SetInput(HDD_LED);
	IO_Activate(HDD_RST);
	mdelay(100);
	IO_Deactivate(HDD_RST);
	IO_Activate(HDD_BUF_EN);
	mdelay(100);
	IDE_WaitForBusyCleared();
}

static void IDE_Off(void)
{
	IO_Deactivate(HDD_BUF_EN);
	IO_Activate(HDD_RST);
	IO_Deactivate(HDD_PWR_ON);
	IO_Activate(HDD_LED);
	IO_Deactivate(HDD_IRQ);
	IO_Deactivate(HDD_DRQ);
	IO_Deactivate(HDD_DACK);
}

static int IDE_GetTemperature(void)
{
	rIDEDEVICEHEAD = 0xA0;
	rIDEFEATURES = 0xDF;
	IDE_DoCommand(0xEF);
	return (rIDESECTORNUMBER);
}

void tomtom_hddtemp_setaddress(unsigned long address)
{
	ideBaseAddress = address;
}

int tomtom_hddtemp_check(void)
{
	int temperature;

	IDE_InitDisk();
	temperature = IDE_GetTemperature();
	IDE_Off();
	if ((temperature < 1) || (temperature > 64)) return -1; else return 0;
}

/* EOF */
